
#include <Arduino.h>
#include "piezo.h"


using namespace piezo;


namespace piezo {


/// @brief              The RingBuffer stores value pairs (analog value and milliseconds) in a type of ring buffer. 
///                     It is specially designed for analyzing the AD values of the piezo sensor, i.e. 
///                         a)  When a value is written, it also stores the ms.
///                         b)  The first value of a sequence of values is stored the ms since the start of the CPU, 
///                             but the following values only the delta to the previous one.
///                         c)  The first value that is saved sets a flag so that the application can start a timer.
///                         d)  When a value is added, the actual ms of the CPU is also saved in a global variable.
///                         e)  The values can be read and removed with Get or Pop.
///                         f)  The values can be read with Peek without being removed so you can apply different 
///                             calculation with the same value serie.
///                         g)  The read and write functions return a Boolean with the information “passed / failed”.
///  Title:             Piezo Sensosr Analysing
///@author              Marc Brotbeck
///  EMail:             marc@brotbeck.com
///@date                Jan 2024
///  Environment:       Visual Studio Code with Platform.io and Arduino C++
///  Companyname        TIS ( Technische Informatik Solutions)
///@copyright           2024  by Technische Informatik Solutions
///     
///  File               piezo.cpp with piezo.h
///@version             V1.0-0




     /// Ctor 
    /// @brief          The RingBuffer stores value pairs (AD value and milliseconds) in a type of ring buffer. 
    ///                 It is specially designed for analyzing the AD values of the piezo sensor, i.e. 
    /// @param size     The number of Element the Ringbuffer can store.
    RingBuffer::RingBuffer(int size) {

       _size        = size;
       _buf         = new Value[size];
       _freePos     = 0;
       _firstPos    = 0;
       _msStart     = 0;
       _currentPos  = 0;
       _lastPos     = 0;
       _empty       = true;
        FlVibStart  = false;
        MsVibLast   = 0;
    }

    /// @brief          Clears the Ringbuffer
    void RingBuffer::Clear() {

       _freePos     = 0;
       _firstPos    = 0;
       _msStart     = 0;
       _msDelta     = 0;
       _currentPos  = 0;
       _lastPos     = 0;
       _empty       = true;
        FlVibStart  = false;
        MsVibLast   = 0;
    }


    /// @brief          Increment the Index with round robin
    /// @param i        Ref of index
    void RingBuffer::inc(int* i) {
        *i = ++ * i % _size;
    }

    /// @brief          Decrement the index with round robin
    /// @param i        Ref of index
    void RingBuffer::dec(int* i) {
        *i = *i == 0 ? _size - 1 : --*i;
    }

    /// @brief          Add a Element of struct to the Ringbuffer if he is not full. The value is stored with the 
    ///                 ms from the cpu for the first element and the delta for each additional value.
    /// @param x        Value (the value from the Analog pin) 
    /// @return         true if added and false if is full.
    bool RingBuffer::Add(int x) {

        Value v;
        if (_freePos == _firstPos && !_empty) return false;
       _msDelta    = millis() - MsVibLast;
        MsVibLast  = millis();
        if (_empty) {
           _msStart    = millis();
           _msDelta    = 0;  
            FlVibStart = true;
        }
        v.val = x;
        v.ms  = _msDelta;
       _empty = false;
       _buf[_freePos] = v;
        inc(&_freePos);

        return true;
    }


   /// @brief           Removes a value and put it in the Parameter by ref 
   /// @param val       is the refernec of struct 
   /// @return          true if the value is written
   bool RingBuffer::Get(Value *val) {

      Value v ; 
      if (_freePos == _firstPos  &&  _empty) return false;
      *val = _buf[_firstPos];
      inc(&_firstPos);
      _empty = _firstPos == _freePos;
      return true ;
   }




    /// @brief          Removes a value by Lifo (stack) and put it in the Parameter by ref. the Value
    ///                 is the last entered and the cell is the free for a new Element.
    ///                 the Last is the the previous Element for the next pop. if it is
    ///                 empty the return is false and the value is not written.
    /// @param val      is the reference of struct and is overwritten by the element. 
    /// @return         true if the value is written
    bool RingBuffer::Pop(Value* val) {

        Value v;
        if (_freePos == _firstPos && _empty) return false;
        dec(&_freePos);
        *val = _buf[_freePos];
        _empty = _firstPos == _freePos;
        return true;
    }



    /// @brief          Reads a value by Fifo at the current position and put it in the Parameter by ref 
    ///                 and increments the current position. if it is the last value in the Ring-
    ///                 buffer the return value is false and the current position is set to the 
    ///                 first Element.
    /// @param val      is the reference of struct and is overwritten by the element.
    /// @return         true if the value is written
    bool RingBuffer::Peek(Value* val) {

        Value v;
       
        if (_currentPos == _freePos && (!_empty && _lastPos || _empty)) {
           _currentPos = _firstPos; // reset to first
           _lastPos    = false;
            return false;
        }
       _lastPos = true;
        *val    = _buf[_currentPos];
        inc(&_currentPos);
        return true;
    }




    /// Ctor
    /// @brief          The fibration analysis calculates the possible vibration behavior of the piezo sensor 
    ///                 from the stored values of the ring buffer.The sensor only provides positive values and 
    ///                 a series of values over a short period of time. All values in a series result in an event.
    ///                 There are two kind of calculation, get and peek,whree get removes the value and peek not.
    ///                 For bring the Ringbuffer ready for a new serie of values at least one get is needed or 
    ///                 otherwise call clear. We calculate this event by:
    ///                 a)   Determining the mean value of all values of the event.
    ///                 b)   Determining the sum of all values.
    ///                 c)   Determine the maximum value.
    ///                 d)   Determine the area of the values, where the values represent the x - axis and the 
    ///                      milliseconds represent the y - axis.
    ///                 ==> The result is an object that was passed as a parameter by reference.
    /// @param size     is the number of values the Ringbuffer should be able to store.
    /// @return         the number of measured values that were processed.
    VibAnalyse::VibAnalyse(int size, byte n) {
    
       _ribu = RingBuffer(size);
        nr   = n;
    }

    /// destructor => nothing yet
    VibAnalyse::~VibAnalyse() { }

    // helper for volume  calculation
    int getLarger(int x, int y) { return x > y ? x : y; }
    int get3Angle(int a, int b) { return a * b / 2; }
    int getDelta( int a, int b) { int d; if (a > b) d = a - b; else d = b - a; return d; }

    // for Test only 
    void VibAnalyse::RandomFillRingBuffer(int max) {

        int x,y;
        srand(time(NULL));
        for (int i = 0; i < max; i++) {
            x = rand() % 100+ 10;
            y = rand() % 100 +1;
            _ribu.Add(x);
            delay(y); // without sleep we got no delta of millis
        }
    }

    void VibAnalyse::AddValue(int x, bool pct) {
        if (pct) x = map(x, 0, 1023, 0, 100);
        _ribu.Add(x);
    }

    /// @brief          collecting value from analog until the Serie is finished. This flag
    ///                 is used th start a serie after n sec a timer function will analyse the 
    ///                 he value. the flag is set by add first values to the EMPTY ringbuffer. While the 
    ///                 analyse is running all values are removed. this is requiered.
    /// @return         true if starts a new serie 
    bool VibAnalyse::NewSerie() {

        if (_ribu.FlVibStart) {
            _ribu.FlVibStart = false;
            return true;
        }
        return false;
    }

    /// @brief          Determines the volume and used time over the collected values in the ringbuffer.
    ///                 The result are written into the value struct pased by reference.
    /// @note           The values are removed from the ringbuffer.
    /// @param v        global value pair millis and volume.
    /// @return         the number of calculated values used from ringbuffer
    int VibAnalyse::GetVolumeVibration(Value* v) {

        int i    = 0;
        int last = 0;
        int ms   = 0;
        int x    = 0;
        Value  y;

        while (_ribu.Pop(&y)) {
            i++;
            ms  += y.ms;
            x   += (getLarger(y.val, last) * y.ms) - (get3Angle(getDelta(y.val, last), y.ms));
            last = y.val;
        }
        v->ms  = ms;
        v->val = x;
        return i;
    }


    /// @brief          Calculates the summe and used time over the collected values in the ringbuffer.
    ///                 The result are written into the value struct pased by reference.
    /// @note           The values are removed from the ringbuffer.
    /// @param v        global value pair millis and volume.
    /// @return         the number of calculated values used from ringbuffer
      int VibAnalyse::GetSumVibration(Value* v) {

        int  i = 0;
        int  x = 0;
        int ms = 0;
        Value  y;

        while (_ribu.Get(&y)) {
            i++;
            ms += y.ms;
            x  += y.val;
        }
        v->ms  = ms;
        v->val = x;

        return i;
    }

    /// @brief          Select the max value and used time over the collected values in the ringbuffer.
    ///                 The result are written into the value struct pased by reference.
    ///                 The values are kept in the ringbuffer for further use.
    /// @param v        global value pair millis and volume.
    /// @return         the number of  values the selected is catchted from ringbuffer
     int VibAnalyse::GetPeakVibration(Value* v) {

        int  i = 0;
        int  x = 0;
        int ms = 0;
        Value  y;

        while (_ribu.Get(&y)) {
            i++;
            ms += y.ms;
            x = x > y.val ? x : y.val;
        }
        v->ms = ms;
        v->val = x;

        return i;
    }
  

    /// @brief          Calculates the averige value and used time over the collected values in the ringbuffer.
    ///                 The result are written into the value struct pased by reference.
    /// @param v        global value pair millis and volume.
    /// @return         the number of calculated values used from ringbuffer
      int VibAnalyse::GetAverageVibration(Value* v) {

        int  i = 0;
        int  x = 0;
        int ms = 0;
        Value  y;

        while (_ribu.Get(&y)) {
            i++;
            ms += y.ms;
            x += y.val;
        }
        v->ms = ms;
        v->val = x/i;

        return i;
    }

    /// The values are kept in the ringbuffer for further use.
    int VibAnalyse::PeekVolumeVibration(Value* v) {

        int i    = 0;
        int last = 0;
        int ms   = 0;
        int x    = 0;
        Value y;

        while (_ribu.Peek(&y)) {
            i++;
            ms += y.ms;
            x += (getLarger(y.val, last) * y.ms) - (get3Angle(getDelta(y.val, last), y.ms));
            last = y.val;
        }
        v->ms = ms;
        v->val = x;
        return i;
    }


    /// The values are kept in the ringbuffer for further use.
    int VibAnalyse::PeekSumVibration(Value* v) {

        int  i = 0;
        int  x = 0;
        int ms = 0;
        Value  y;

        while (_ribu.Peek(&y)) {
            i++;
            ms += y.ms;
            x += y.val;
        }
        v->ms = ms;
        v->val = x;

        return i;
    }


    /// The values are kept in the ringbuffer for further use.
    int VibAnalyse::PeekPeakVibration(Value* v) {

        int  i = 0;
        int  x = 0;
        int ms = 0;
        Value  y;

        while (_ribu.Peek(&y)) {
            i++;
            ms += y.ms;
            x = x > y.val ? x : y.val;
        }
        v->ms = ms;
        v->val = x;

        return i;
    }


    /// The values are kept in the ringbuffer for further use.
   int VibAnalyse::PeekAverageVibration(Value* v) {

        int  i = 0;
        int  x = 0;
        int ms = 0;
        Value  y;

        while (_ribu.Peek(&y)) {
            i++;
            ms += y.ms;
            x += y.val;
        }
        
        v->ms = ms;
        v->val = i == 0 ? 0 : x / i;
        return i;
    }
}
